# -*- coding: UTF-8 -*-
import os
import sys
from itertools import groupby

from common import logger, filesmodify
from lxml import etree as ET

saleFiles = set()


class SmagXmlUnloadConverter(object):
    XML_VERSION = "1.2"

    payment_mapper = {}

    artix_param_to_xml_tag = {
        "articul": "article",
        "timeEnd": "receiptDateTime",
        "timeBeg": "receiptOpenDateTime",
        "docType": "type",
        "sumSale": "saleAmount",
        "sumrefund": "returnAmount",
        "countsale": "saleCount",
        "countrefund": "returnCount",
        "userCode": "cashierCode",
        "userName": "cashierName",
        "barCode": "barcode",
        "posNum": "position",
        "deptCode": "stockId",
        "bcodeMode": "enterType",
        "posSum": "total",  # Стоимость позиции, уменьшенная на величину скидок
        "markingcode": "KIZ",
        "excisemark": "egaisBarcode",
        "quant": "quantity",
        "price": "price",
        "docSum": "amount",
        "sourceidentifier": "receiptNumSale",
        "fdnumber": "fiscalDocumentNumber",
        "fnnumber": "fnNumber"
    }

    payment_name = {
        0: "Наличные",
        101: "Банковская карта",
        102: "Подарочный сертификат",
        103: "Оплата баллами",
        104: "Предоплата",
        105: "Кредит"
    }

    artix_doctype_to_ukm_doctype = {
        0: "0",
        1: "0",  # продажа
        2: "1",  # возврат
        25: "4",  # возврат по исходному чеку
        4: "6"  # расход (выем)
    }

    def __init__(self, node, work_dir):
        self.log = logger.getUnloadLogger("Node", node.id)
        self.work_dir = work_dir
        self.node = node
        self.shopcode = node.shopCode
        SmagXmlUnloadConverter.payment_mapper = node.currentConverter.paymentMapper

    def run(self, salesFunc, shiftsFunc, cash, currentSales):
        self.log.debug('shopcode=%s cashcode=%s' % (self.shopcode, cash.code))
        shifts = {}
        shiftsNotOpen = []
        res = shiftsFunc()
        shift_files = set()
        for sh in res:
            if sh['timeBeg'] is None:
                shiftsNotOpen.append(sh['shift'])
            else:
                shifts[sh['shift']] = sh
        del res
        if shifts:
            self.log.info(u"Выгрузка продаж для смен: %s" % shifts.keys())
            self.log.info(
                u"Не закрытые и сторнированные чеки, а также чеки без товарных позиций не будут выгружены!")
        if currentSales:
            self.__create_documents_xml(salesFunc(), self.shopcode, cash.code, shifts)
        else:
            for num, docs in groupby(salesFunc(), key=lambda x: x["shift"]):
                if not self.node.currentConverter.unloadNotOpenedShift and (num in shiftsNotOpen):
                    continue
                documents = (d for d in docs)
                del docs
                if not documents and not self.node.currentConverter.withClosedShifts:
                    self.log.debug("Смена %s не содержит документов" % num)
                    continue
                shift_file = self.__create_shift_xml(self.shopcode, cash.code, shifts[num], documents)
                shift_files.add(shift_file)
                del shift_file
                del documents
            self.__post_unload(self.work_dir, False, shift_files)
        for item in shifts:
            del item
        del shifts

    def __create_shift_xml(self, shopcode, cashcode, shift, documents):
        """
        сборка файла xml для смены
        """
        shiftnum = shift['shift']
        shift_filename = "/%s/shift_[%s]_[%s]_[%s]_[1].xml" % (self.work_dir, shopcode, cashcode, shiftnum)
        self.log.info("Смена будет выгружена в файл  %s" % shift_filename)
        with ET.xmlfile(shift_filename, encoding='utf-8', close=True) as xf:
            xf.write_declaration()
            with xf.element('shift', storeId=shopcode, posNum=cashcode, shiftNum=str(shiftnum)):
                # перевод каретки для форматирования
                xf.write("\n")
                # версия xml
                self.__write_et(self.__get_version(), xf)
                # дата закрытия смены
                xml_elem = self.__create_etree_element("dateClose", self.__get_formatted_date(shift["timeEnd"]))
                self.__write_et(xml_elem, xf)
                # код и имя кассира
                user_data = self.__get_user_ET(shift["userCode"], shift["users"])
                self.__write_et(user_data, xf)
                # счетчики смены
                for key in ["sumSale", "sumrefund", "countsale", "countrefund"]:
                    xml_elem = self.__create_etree_element(SmagXmlUnloadConverter.artix_param_to_xml_tag[key],
                                                           shift[key])
                    self.__write_et(xml_elem, xf)
                # данные ккм
                kkm_xml_elements = self.__get_kkm_ET(shift["kkms"])
                self.__write_et(kkm_xml_elements, xf)
                # оплаты в смене, их немного, поэтому можно грузить списком
                shift_payments = self.__get_payment_ET(shift)
                self.__write_et(shift_payments, xf)
                # чеки
                written_check_counter = 0
                for check in documents:
                    if not self.__validate_check(check):
                        continue
                    # это для форматирования
                    if written_check_counter > 0:
                        xf.write("\n")
                    self.__create_and_write_documents_etree(shopcode, cashcode, shift, check, xf)
                    written_check_counter += 1
                    del check
                del documents
                self.log.debug("Выгружено чеков: %s" % written_check_counter)
            del xf
        return shift_filename

    def __create_documents_xml(self, documents, shopcode, cashcode, shifts):
        self.log.info("Выгрузка оперативных продаж, магазин %s, касса %s" % (shopcode, cashcode))
        check_files = set()
        for check in documents:
            if not self.__validate_check(check):
                continue
            check_filename = "/%s/receipt_[%s]_[%s]_[%s]_[%s]_[1].xml" % (self.work_dir,
                                                                          shopcode, cashcode, check["shift"],
                                                                          check["docNum"])
            self.log.info("Чек будет выгружен в файл %s" % check_filename)
            with ET.xmlfile(check_filename, encoding='utf-8', close=True) as xf:
                xf.write_declaration()
                self.__create_and_write_documents_etree(shopcode, cashcode, shifts[check["shift"]], check, xf)
                del check
                check_files.add(check_filename)
        self.__post_unload(self.work_dir, True, check_files)
        del documents
        for sh in shifts:
            del sh
        del shifts

    def __create_and_write_documents_etree(self, shopcode, cashcode, shift, check, xf):
        # чеки
        checknum = check["docNum"]
        with xf.element("receipt", storeId=shopcode, posNum=cashcode, shiftNum=str(check["shift"]),
                        receiptNum=str(checknum)):
            xf.write("\n")
            # версия вроде как не нужна в чеке [так передали клиенты 7 шагов]
            # self._write_et(self._get_version(), xf)
            # даты в чеке
            self.__write_et(self.__get_check_datetime_ET(check["timeBeg"], check["timeEnd"]), xf)
            # имя и код кассира
            self.__write_et(self.__get_user_ET(check["userCode"], shift["users"]), xf)
            # doctype
            self.__write_et(self.__get_check_doctype_xml(check["docType"]), xf)
            # sumb
            self.__write_et(self.__create_etree_element("docSum", abs(check["docSum"])), xf)
            if check["docType"] in [2, 25]:
                self.__write_et(self.__create_etree_element("sourceidentifier", check["sourceidentifier"]), xf)
            # собираем позиции чека
            for invent in check["inventPositions"]:
                item = self.__create_invent_ET(invent)
                self.__write_et(item, xf)
            # собираем оплаты чека
            moneyitems = check["moneyPositions"]
            # рассматриваем случай оплаты со сдачей - тут нам нужно вычислить только ту оплату,
            # где действительно приняли деньги от клиента
            # надежнооо, как швейцарские часы
            if len(moneyitems) == 3 and len(list(filter(lambda m: m["valutOperation"] == 0, moneyitems))) == 3:
                reverse_payments = list(filter(lambda m: m["opCode"] == 72, moneyitems))
                if len(reverse_payments) == 1:
                    original_payments = list(
                        filter(lambda m: m["opCode"] == 70 and m["sumB"] != reverse_payments[0]["sumB"], moneyitems))
                    if len(original_payments) == 1:
                        moneyitem_et = self.__create_moneyitem_ET(original_payments[0])
                        self.__write_et(moneyitem_et, xf)
            else:
                for payment in moneyitems:
                    moneyitem_et = self.__create_moneyitem_ET(payment)
                    self.__write_et(moneyitem_et, xf)
            del moneyitems
            # заполняем фискальные признаки
            if len(check["fiscaldocument"]) > 0:
                fiscaldoc = check["fiscaldocument"][0]
                self.__write_et(self.__create_etree_element("fdnumber", fiscaldoc["fdnumber"]), xf)
                self.__write_et(self.__create_etree_element("fnnumber", fiscaldoc["fnnumber"]), xf)
                del fiscaldoc
            xf.flush()
            del check

    @staticmethod
    def __post_unload(work_dir, is_current_sales, shift_files):
        statusFileName = "ready"
        statusDir = work_dir
        if is_current_sales:
            statusFileName += "Online"
            statusDir = os.path.dirname(work_dir)
        converterSalesReadyFile = open('%s/%s' % (statusDir, statusFileName), 'w')
        converterSalesReadyFile.close()
        salesReadyFile = open('%s/%s' % (os.path.dirname(os.path.dirname(statusDir)), statusFileName), 'w')
        salesReadyFile.close()
        shift_files.add(converterSalesReadyFile.name)
        shift_files.add(salesReadyFile.name)
        global saleFiles
        saleFiles = shift_files
        del shift_files

    @staticmethod
    def __create_etree_element(key, text, parent_node=None, attrib=None):
        """
        создание xml-элемента
        """
        if text is None or text == "":
            return None
        if attrib is None:
            attrib = {}
        tag = SmagXmlUnloadConverter.artix_param_to_xml_tag.get(key, key)
        elem = ET.Element(tag, **attrib) if parent_node is None else ET.SubElement(parent_node, tag, **attrib)
        elem.text = SmagXmlUnloadConverter.__get_xml_text(text)
        return elem

    @staticmethod
    def __create_etree_elements(data_dict, parent_node=None):
        """
        создание дочерних xml-тегов из словаря для родительской ноды
        """
        elems = [SmagXmlUnloadConverter.__create_etree_element(tag, text, parent_node) for tag, text in
                 list(data_dict.items())] if data_dict is not None else []
        return parent_node if parent_node is not None else elems if len(elems) > 0 else None

    @staticmethod
    def __get_xml_text(text):
        """
        костыль для поддержки python2 - строка с кириллицей для записи в xml должна быть типа unicode
        """
        if sys.version_info.major == 3:
            return str(text)
        else:
            # из кириллицы нужно сделать unicode
            if isinstance(text, unicode):
                return text
            elif isinstance(text, str):
                return unicode(text, "utf-8")
            else:
                # числа оборачиваем в строку
                return str(text)

    @staticmethod
    def __write_et(et_xml, xmlfile):
        """
        запись в файл списка или элемента с форматированием
        """
        # пишем только заполненные теги
        if isinstance(et_xml, list):
            for el in et_xml:
                xmlfile.write(el, pretty_print=True)
                xmlfile.flush()
                del el
        else:
            xmlfile.write(et_xml, pretty_print=True)
            xmlfile.flush()
            del et_xml

    @staticmethod
    def __get_version():
        """
        версия укм
        """
        return SmagXmlUnloadConverter.__create_etree_element("version", SmagXmlUnloadConverter.XML_VERSION)

    @staticmethod
    def __get_check_doctype_xml(docType):
        """
        тип документа
        """
        return SmagXmlUnloadConverter.__create_etree_element("docType",
                                                             SmagXmlUnloadConverter.artix_doctype_to_ukm_doctype[
                                                                 docType]) if docType in SmagXmlUnloadConverter.artix_doctype_to_ukm_doctype else None

    @staticmethod
    def __get_formatted_date(date):
        """
        форматирование даты под нужный формат
        """
        return date.strftime("%Y-%m-%dT%H:%M:%S") if date is not None else ""

    @staticmethod
    def __get_user_ET(usercode, users):
        """
        код и имя кассира
        """
        users_dict = {user["usercode"]: user for user in users}
        name = users_dict.get(usercode)["username"] if usercode in users_dict.keys() else ""
        name = name if sys.version_info.major == 3 else name.encode("UTF-8")
        return [SmagXmlUnloadConverter.__create_etree_element("cashierCode", usercode),
                SmagXmlUnloadConverter.__create_etree_element("cashierName", name)]

    @staticmethod
    def __get_kkm_ET(kkms):
        """
        получить данные по ккм смены
        TODO: на данный момент не рассматриваем случай нескольких ФР у клиента, берем первый элемент из списка
        """
        result = dict()
        if len(kkms) > 0:
            kkm = kkms[0]
            result["kkm_shift_number"] = kkm["shiftnumkkm"]
            result["kkm_serial_number"] = kkm["serialNum"]
            result["kkm_registration_number"] = kkm["registrationnumber"]
            result["kkm_model_name"] = kkm["modelName"]
            result["kkm_saleAmount"] = kkm["sumSale"]
            result["kkm_returnAmount"] = kkm["sumBack"]
            result["kkm_saleCount"] = kkm["countSale"]
            result["kkm_returnCount"] = kkm["countBack"]
            result["kkm_withdrawal"] = kkm["countMoneyOut"]
            result["kkm_insertion"] = kkm['countMoneyIn']
        return SmagXmlUnloadConverter.__create_etree_elements(result)

    @staticmethod
    def __get_payment_ET(shift):
        """
        xml-элементы для оплат в смене, тег <payment>
        """
        payment_cash = SmagXmlUnloadConverter.__create_payment(0, shift)
        payment_non_cash = SmagXmlUnloadConverter.__create_payment(101, shift)
        return [SmagXmlUnloadConverter.__create_etree_elements(payment, ET.Element("payment")) for payment in
                [payment_cash, payment_non_cash]]

    @staticmethod
    def __create_payment(paymentId, shift):
        """
        формируем данные по оплатам, на данный момент только наличные/карта
        """

        payment = dict()
        payment["paymentId"] = paymentId
        name = SmagXmlUnloadConverter.payment_name.get(paymentId, "Неизвестно")
        payment["paymentName"] = name
        payment["paymentAmountSale"] = shift["sumsalecash"] if paymentId == 0 else shift["sumsalenoncash"]
        payment["paymentAmountReturn"] = shift["sumrefundcash"] if paymentId == 0 else shift["sumrefundnoncash"]
        if payment["paymentAmountSale"] == 0.00 and payment["paymentAmountReturn"] == 0.00:
            return None
        # payment["paymentAmountExpense"] = ""  # Нет описания, не заполняем
        # ищем код типа оплаты артикс по paymentId ukm
        payment_map = SmagXmlUnloadConverter.payment_mapper
        artix_payment_type = 0
        if str(paymentId) in payment_map.values():
            artix_payment_type = list(payment_map.keys())[list(payment_map.values()).index(str(paymentId))]
        if len(shift["kkms"]) > 0:
            kkm = shift["kkms"][0]
            payment["kkm_paymentAmountSale"] = kkm.get("sumSaleByPayment%s" % artix_payment_type, "0.00")
            payment["kkm_paymentAmountReturn"] = kkm.get("sumBackByPayment%s" % artix_payment_type, "0.00")
        # payment["kkmPaymentAmountExpense"] = ""  # Нет описания, не заполняем
        return payment

    @staticmethod
    def __create_invent_ET(goods_item):
        """
        xml-элементы для товара, тег <item>
        """

        # маппинг способов ввода товара
        artix_input_mode_to_ukm = {
            "manual": "1",  # вручную -> клавиатура
            "scanner": "0",  # сканером(2), из софт-чека(9), через распознавание(10) -> 0 – сканер, заказ, распознавание
            "goodsdetector": "0",
            "softCheck": "0",
            "hotkey": "2",
            # по горячей клавише(3), при выборе из списка (по коду, по штрих-коду, по артикулу, по наименованию,
            # по цене) (4)
            "selected": "2"
        }

        item_root_tag = ET.Element("item")
        item = dict()
        item["position"] = goods_item["posNum"]
        item["article"] = goods_item["articul"]
        item["barcode"] = goods_item["barCode"]
        item["quantity"] = goods_item["quant"]
        item["price"] = goods_item["price"]
        item["total"] = goods_item["posSum"]
        item["stockId"] = goods_item["deptCode"]
        item["enterType"] = artix_input_mode_to_ukm.get(goods_item['bcodeMode'], "10")
        item["KIZ"] = goods_item["markingcode"]
        # У нас нет, надо подумать чем заполнять, возможно excisemark целиком
        item["serialNumber"] = goods_item["excisemark"]
        item["egaisBarcode"] = goods_item["excisemark"]
        # дополняем корневую ноду элементами
        SmagXmlUnloadConverter.__create_etree_elements(item, item_root_tag)
        # налог
        """
        ######################################################## 
        Пока что налог не выгружаем, клиентам вроде как не нужно
        ######################################################## 
        invent_taxgroup_code = goods_item["vatcode1"]
        invent_taxes = list(filter(lambda t: t["taxCode"] == invent_taxgroup_code, goods_item["taxes"]))
        tax = dict()
        if len(invent_taxes) == 1:
            invent_tax = invent_taxes[0]
            tax["TaxGroupId"] = int((int(invent_taxgroup_code)-1)/100)
            tax["TaxId"] = invent_taxgroup_code
            tax["percent"] = invent_tax["taxRate"]
            tax["amount"] = invent_tax["taxSum"]
        item_root_tag.append(SmagXmlUnloadConverter.__create_etree_elements(tax, ET.Element("Tax")))
        """
        # скидки
        for disc in goods_item["discountPositions"]:
            discount = dict()
            discount["programId"] = disc["discCode"]
            discount["programName"] = disc["discName"]
            # в доке клиентов programType должно быть числом, пока не заполняем
            #discount["programType"] = disc["discType"]
            discount["amount"] = disc["discSum"]
            discount["discCardNumber"] = disc["cardNumber"]
            item_root_tag.append(SmagXmlUnloadConverter.__create_etree_elements(discount, ET.Element("discount")))
        return item_root_tag

    def __create_moneyitem_ET(self, data):
        """
        xml-элементы для оплат в чеке, тег <payment>
        """
        # маппинг типов оплаты по полю valutoperation
        artix_payment_type_to_ukm = {
            0: "0",  # Наличные
            10: "101",  # Банковская карта
            3: "103"  # Подарочный сертификат
            # 104 – аванс/предоплата
            # 105 – оплата в кредит
        }

        payment_root_tag = ET.Element("payment")
        moneyitem = dict()
        if data['valutOperation'] not in artix_payment_type_to_ukm:
            self.log.info("Выгрузка оплаты с типом valutoperation=%s не поддержана" % data['valutOperation'])
            return
        pay_id = artix_payment_type_to_ukm[data['valutOperation']]
        moneyitem["paymentId"] = pay_id
        moneyitem["paymentName"] = SmagXmlUnloadConverter.payment_name[int(pay_id)]
        moneyitem["paymentAmount"] = data['sumB']
        moneyitem["maskedCardNumber"] = data['cardnum']
        moneyitem["cardType"] = data['docnum']
        moneyitem["authCode"] = data['authcode']
        moneyitem["rrn"] = data['sourceoperationid']
        # дополняем корневую ноду элементами
        SmagXmlUnloadConverter.__create_etree_elements(moneyitem, payment_root_tag)
        return payment_root_tag

    @staticmethod
    def __get_check_datetime_ET(timeBeg, timeEnd):
        """
        даты в чеке
        """
        timeBeg = SmagXmlUnloadConverter.__create_etree_element("timeBeg",
                                                                SmagXmlUnloadConverter.__get_formatted_date(timeBeg))
        timeend = SmagXmlUnloadConverter.__get_formatted_date(timeEnd)
        timeEnd = SmagXmlUnloadConverter.__create_etree_element("timeEnd", timeend)
        subTotal = SmagXmlUnloadConverter.__create_etree_element("receiptSubtotalDateTime", timeend)

        return [timeBeg, timeEnd, subTotal]

    def __validate_check(self, check):
        docType = check["docType"]
        checknum = check["docNum"]
        if docType not in self.artix_doctype_to_ukm_doctype:
            self.log.debug(
                "Работа с типом чека %s не поддержана, чек №%s не будет выгружен" % (docType, checknum))
            return False
        if docType == 4:
            self.log.debug(
                "Чек №%s является чеком выема, на данный момент выгрузка такого чека не поддержана" % (checknum))
            return False
        if check["closed"] == 2:
            self.log.debug("Чек №%s был сторнирован и не будет выгружен" % checknum)
            return False
        if len(check["inventPositions"]) == 0:
            self.log.debug("Чек №%s не содержит товарных позиций и не будет выгружен" % checknum)
            return False
        if check["closed"] != 1:
            self.log.debug("Чек №%s не был закрыт и не будет выгружен" % checknum)
            return False
        return True


def run(salesFunc, securityLogsFunc, shiftsFunc, cash, node, workDirectory, currentSales):
    SmagXmlUnloadConverter(node, workDirectory).run(salesFunc, shiftsFunc, cash, currentSales)


def sendOnlineSales(cash, node, workDirectory):
    log = logger.getUnloadLogger("Node", node.id)
    return filesmodify.safeMoveDirectory(log, workDirectory, os.path.dirname(node.currentConverter.onlineSalePath))


def getSaleFiles():
    global saleFiles
    return saleFiles


def sendSales(cash, node, workDirectory, backupDir):
    log = logger.getUnloadLogger("Node", node.id)
    return filesmodify.safeMoveDirectory(log, workDirectory, os.path.dirname(node.currentConverter.path),
                                         rewrite=node.currentConverter.rewriteDestFile, backupDir=backupDir,
                                         maxCountFiles=cash.maxCountFiles, actualCountDays=cash.actualCountDays)


def parseFlag(flag):
    f = open(flag, 'r')
    shifts = None
    dates = None
    if f.readline().strip() == '###':
        str = f.readline().strip()
        if len(str.split('-')) == 2:
            if str.find('.') == -1:
                shifts = str.replace('-', '@')
            else:
                fromDate = str.split('-')[0].split('.')
                toDate = str.split('-')[1].split('.')
                dates = '%s-%s-%sT00:00:00.0@%s-%s-%sT23:59:59.99' % (
                    fromDate[2], fromDate[1], fromDate[0], toDate[2], toDate[1], toDate[0])
    f.close()
    return shifts, dates
