# -*- coding: UTF-8 -*-

# cif1 - custom interchange format 1

import os
import struct
from common import aif_objects as aif, logger
from upload.inserter import InserterError

errorObjectsCount = 0

def vbaToUnixTimestamp(vbaTimestamp):
    return round((vbaTimestamp * 86400) - 2209161600)

def filesListWithExtension(extension, files):
    return [fileName for fileName in files if fileName.endswith(extension)]

def trimString(string):
    return string.split(b'\0', 1)[0]

class Converter(object):
    def __init__(self, workDirectory, data, node):
        self.log = logger.getUploadLogger("Node", node.id)
        self.workDirectory = workDirectory
        self.data = data
        self.node = node
        plugin = __import__('upload.inserter.%s' % node.inserter.name, fromlist='Inserter')
        cl = getattr(plugin, 'Inserter')
        self.inserter = cl(node)
        self.listObjects = []
        self.taxes = []

    def appendDict(self, aifObject):
        self.listObjects.append(aifObject)
        self.listObjects = list(filter(None, self.listObjects))
        if len(self.listObjects) >= self.node.pack_size:
            self.inserter.add_package(self.listObjects)
            self.listObjects = []

    def run(self):
        global errorObjectsCount

        self.prepare()

        result = self.inserter.start()
        for brcFile in filesListWithExtension('.brc', self.data):
            self.parseBarcodesBrcFormat(brcFile)
        for barFile in filesListWithExtension('.bar', self.data):
            self.parseBarcodesBarFormat(barFile)
        for prdFile in filesListWithExtension('.prd', self.data):
            self.parseInventPrdFormat(prdFile)
        for sprFile in filesListWithExtension('.spr', self.data):
            self.parseInventSprFormat(sprFile)

        self.postpare()

        result = self.inserter.stop()
        errorObjectsCount = self.inserter.invalidObjectsCount()

        return result

    def prepare(self):
        '''
        Функция подготовки данных для загрузки справочников
        '''
        self.appendDict({'command': 'clearUnit'})
        self.appendDict({'command': 'clearLVat'})
        self.appendDict({'command': 'clearSaleRestrict'})
        self.appendDict({'command': 'clearInventory'})

        self.appendDict(aif.Unit().addCmd(unitcode=1, name=u'шт', fractional=False))
        self.appendDict(aif.Unit().addCmd(unitcode=2, name=u'кг', fractional=True))

    def postpare(self):
        for tax in self.taxes:
            self.listObjects.append(tax)
        self.listObjects = list(filter(None, self.listObjects))
        self.inserter.add_package(self.listObjects)
        self.listObjects = []

    def parseBarcodesBrcFormat(self, url):
        '''
        Функция создает aif-объекты штрихкодов из справочника штрихкодов и добавляет данные в inserter
        '''
        fileSize = os.path.getsize(url)
        with open(url, 'rb') as barcodeFile:
            currentFilePosition = 0
            while currentFilePosition < fileSize:
                nodeSize = struct.unpack('=H', barcodeFile.read(2))[0]
                inventCode, defaultQuantity, barcode = self.parseBarcodeBrcFormat(barcodeFile)
                self.appendDict(aif.Barcode().addCmd(code=inventCode, barcode=barcode, quantdefault=defaultQuantity))
                currentFilePosition += nodeSize + 2
                barcodeFile.seek(currentFilePosition, 0)

    def parseBarcodesBarFormat(self, url):
        '''
        Функция создает aif-объекты штрихкодов из справочника штрихкодов и добавляет данные в inserter
        '''
        fileSize = os.path.getsize(url)
        with open(url, 'rb') as barcodeFile:
            while barcodeFile.tell() < fileSize:
                inventCode, defaultQuantity, barcode = self.parseBarcodeBarFormat(barcodeFile)
                self.appendDict(aif.Barcode().addCmd(code=inventCode, barcode=barcode, quantdefault=defaultQuantity))

    def parseInventPrdFormat(self, url):
        fileSize = os.path.getsize(url)
        with open(url, 'rb') as tmcFile:
            currentFilePosition = 0
            # цикл по нодам
            while currentFilePosition < fileSize:
                nodeSize = struct.unpack('=H', tmcFile.read(2))[0]
                self.parseNode(tmcFile, nodeSize)
                currentFilePosition += nodeSize + 2 # 2 - размер поля размера данных
                tmcFile.seek(currentFilePosition, 0)

    def parseInventSprFormat(self, url):
        fileSize = os.path.getsize(url)
        with open(url, 'rb') as tmcFile:
            self.parseNode(tmcFile, fileSize)

    def parseNode(self, tmcFile, nodeSize):
        # проход по ноде
        currentFilePosition = tmcFile.tell()
        nodeBeginPosition = tmcFile.tell()

        aifInventAddCmd = None
        minPrice = None
        restrictionsList = []
        while tmcFile.tell() < nodeBeginPosition + nodeSize:
            # считать код поля и длину поля
            structInfo = struct.unpack('=BH', tmcFile.read(3))
            try:
                if structInfo[0] == 0:
                    aifInventAddCmd = self.parseTKod0(tmcFile)
                elif structInfo[0] == 1:
                    self.parseTKod1(tmcFile)
                elif structInfo[0] == 2:
                    self.parseTKod2(tmcFile)
                elif structInfo[0] == 3:
                    self.parseTKod3(tmcFile)
                elif structInfo[0] == 4:
                    restrictionsList.extend(self.parseTKod4(tmcFile))
                elif structInfo[0] == 5:
                    self.parseTKod5(tmcFile)
                elif structInfo[0] == 6:
                    # в справочнике может быть несколько минимальных цен,
                    # но так как у нашего товара она может быть одна,
                    # то в товар запишется значение последней считанной минимальной цены
                    minPrice = self.parseTKod6(tmcFile)
                elif structInfo[0] == 7:
                    self.parseTKod7(tmcFile)
                elif structInfo[0] == 8:
                    self.parseTKod8(tmcFile)
                elif structInfo[0] == 9:
                    self.parseTKod9(tmcFile)
                elif structInfo[0] == 10:
                    self.parseTKod10(tmcFile)
                elif structInfo[0] == 17:
                    self.parseTKod17(tmcFile)
                elif structInfo[0] == 20:
                    self.parseTKod20(tmcFile)
                elif structInfo[0] == 21:
                    self.parseTKod21(tmcFile)
                elif structInfo[0] == 22:
                    self.parseTKod22(tmcFile)
                elif structInfo[0] == 23:
                    self.parseTKod23(tmcFile)
                elif structInfo[0] == 24:
                    self.parseTKod24(tmcFile)
                elif structInfo[0] == 25:
                    self.parseTKod25(tmcFile)
                elif structInfo[0] == 30:
                    self.parseTKod30(tmcFile)
                elif structInfo[0] == 31:
                    self.parseTKod31(tmcFile)
                elif structInfo[0] == 32:
                    self.parseTKod32(tmcFile)
                elif structInfo[0] == 33:
                    self.parseTKod33(tmcFile)
                elif structInfo[0] == 34:
                    self.parseTKod34(tmcFile)
                elif structInfo[0] == 35:
                    self.parseTKod35(tmcFile)
                elif structInfo[0] == 50:
                    self.parseTKod50(tmcFile)
                else:
                    raise Exception(u'Нет правил чтения для структуры с номером %d' % structInfo[0])
            except Exception as e:
                for errorLine in e.args:
                    self.log.error(errorLine)
                    errorList.append(errorLine)

            currentFilePosition += structInfo[1] + 3 # размер информации о структуре
            tmcFile.seek(currentFilePosition, 0)
        if aifInventAddCmd and minPrice:
            aifInventAddCmd['invent']['minprice'] = minPrice
        if aifInventAddCmd:
            aifInventAddCmd['invent']['sellrestrictperiods'] = restrictionsList
        self.appendDict(aifInventAddCmd)

    def parseTKod0(self, tmcFile):
        inventCode = str(struct.unpack('=L', tmcFile.read(4))[0])
        inventNameLength = struct.unpack('=B', tmcFile.read(1))[0]

        # согласно протоколу: если у товара нет наименования, то остальные атрибуты тоже отсутствуют
        if inventNameLength == 0:
            raise Exception(u'Товар не имеет атрибутов')

        inventName = trimString(tmcFile.read(inventNameLength).decode('cp1251').encode('utf8'))
        inventPrice = struct.unpack('=d', tmcFile.read(8))[0]

        inventTax = struct.unpack('=d', tmcFile.read(8))[0] # налоговая ставка
        inventSalesTax = struct.unpack('=d', tmcFile.read(8))[0] # налог с продаж
        inventIsFractional = struct.unpack('=B', tmcFile.read(1))[0] # узнать каким образом грузятся справочники - с очисткой или без
        inventStampLength = struct.unpack('=B', tmcFile.read(1))[0]
        inventStamp = trimString(tmcFile.read(inventStampLength).decode('cp1251').encode('utf8')) # какая-то марка в extendetoptions?
        vbaTimestamp = struct.unpack('=d', tmcFile.read(8))[0] # дата товара
        inventType = struct.unpack('=B', tmcFile.read(1))[0] # 0 - обычный, 1 - алкогольный, 2 - табак
        inventValutType = struct.unpack('=B', tmcFile.read(1))[0] # вид оплаты - не используется
        inventAlcoFlag = struct.unpack('=B', tmcFile.read(1))[0] # флаг слабого алкоголя(0 - прочее, 1 - слабый алкоголь)
        inventMultipackQuantity = struct.unpack('=H', tmcFile.read(2))[0] # количество для мультипака - пусть будет количеством по умолчанию

        options = {}
        options['inventitemoptions'] = {}
        options['inventitemoptions']['egaisverify'] = 1 if (inventType == 1 and inventAlcoFlag == 0) else 0
        options['inventitemoptions']['ageverify'] = 1 if inventType == 2 else 0
        #options['inventitemoptions']['nopdfegaisverify'] = 1 if inventAlcoFlag == 1 else 0
        return aif.Invent().addCmd(
            inventcode=inventCode,
            barcode=None,
            name=inventName.decode(),
            price=inventPrice,
            measurecode=2 if inventIsFractional else 1,  # захардкожено!
            extendedoptions=inventStamp.decode(),
            lvatcode=[self.getTaxCode(inventTax), 0, 0, 0, 0],
            defaultquantity=inventMultipackQuantity,
            options=options)

    def parseTKod1(self, tmcFile):
        '''
        Скидка интервальная - работает с hh1:mm1:ss1 до hh2:mm2:ss2 по дням - maska
        '''
        values = struct.unpack('=HHBdBBH', tmcFile.read(17))
#        print('TKod1:')
#        print('  VremahNach = %d' % values[0]) # узнать про преобразование
#        print('  VremahKon = %d' % values[1]) # узнать про преобразование
#        print('  Maska = %d' % values[2])
#        print('  Cena = %f' % values[3])
#        print('  KodKarty = %d' % values[4])
#        print('  Kod = %d' % values[5])
#        print('  VidAkc = %d' % values[6])

    def parseTKod2(self, tmcFile):
        '''
        Скидка по картам
        '''
        values = struct.unpack('=BBdBH', tmcFile.read(13))

    def parseTKod3(self, tmcFile):
        '''
        Акция
        '''
        values = struct.unpack('=BLBBHLddBH', tmcFile.read(32))

    def getRestrict(self, startHour, startMinute, endHour, endMinute, daystart, dayend):
        '''
        Получить ограничение продажи
        Если время начала больше времени окончания,
        то разбиваем на 2 ограничения
        '''
        timestart = '%s:%s:00' % (str(startHour).rjust(2, '0'), str(startMinute).rjust(2, '0'))
        timeend = '%s:%s:00' % (str(endHour).rjust(2, '0'), str(endMinute).rjust(2, '0'))

        result = []
        if startHour > endHour:
            result.append({'timestart': timestart, 'timeend': '23:59:59', 'daystart': daystart, 'dayend': dayend})
            result.append({'timestart': '00:00:00', 'timeend': timeend, 'daystart': daystart, 'dayend': dayend})
        else:
            result.append({'timestart': timestart, 'timeend': timeend, 'daystart': daystart, 'dayend': dayend})

        return result

    def parseTKod4(self, tmcFile):
        '''
        Запрет продажи
        Возвращает список aif-объектов SaleRestrict
        '''
        values = struct.unpack('=BBBBB', tmcFile.read(5))
        startHour = values[0]
        startMinute = values[1]
        endHour = values[2]
        endMinute = values[3]
        dayMask = values[4]
        restrictionsList = []

        first_day = None
        last_day = None

        for day in range(0, 7):
            """
            Накапливаем инф-ию о ограничениях, чтобы записать в БД 1 или 2 строки
            Например пн-вс 22:00 до 8:00 в БД будет как
            1-7 22:00-23:59
            1-7 00:00-08:00
            """
            if (1 << day) & dayMask:
                if not first_day:
                    first_day = day + 1
                last_day = day + 1
            else:
                if not first_day or not last_day:
                    continue

                restrictionsList.extend(self.getRestrict(startHour, startMinute, endHour, endMinute, first_day, last_day))
                first_day = None
                last_day = None

        if first_day or last_day or not restrictionsList:
            restrictionsList.extend(self.getRestrict(startHour, startMinute, endHour, endMinute, first_day, last_day))

        return restrictionsList

    def parseTKod5(self, tmcFile):
        '''
        Список тары
        '''
        values = struct.unpack('=d', tmcFile.read(8))

    def parseTKod6(self, tmcFile):
        '''
        Минимальная цена
        '''
        values = struct.unpack('=d', tmcFile.read(8))
        return values[0]

    def parseTKod7(self, tmcFile):
        '''
        Уценка
        '''
        values = struct.unpack('=LLLBHH', tmcFile.read(17))

    def parseTKod8(self, tmcFile):
        '''
        Цена по штрихкоду - акция
        '''
        values = struct.unpack('=LLd', tmcFile.read(16))
        barcodeLength = struct.unpack('=B', tmcFile.read(1))[0]
        barcode = trimString(tmcFile.read(barcodeLength).decode('cp1251').encode('utf8'))

    def parseTKod9(self, tmcFile):
        '''
        Реверсная акция
        '''
        values = struct.unpack('=H', tmcFile.read(2))

    def parseTKod10(self, tmcFile):
        '''
        Скидка интервальная
        '''
        values = struct.unpack('=HHBdBHH', tmcFile.read(18))

    def parseTKod17(self, tmcFile):
        '''
        Уценка тухлятины
        '''
        values = struct.unpack('=BBHBBHdHH', tmcFile.read(20))

    def parseTKod20(self, tmcFile):
        '''
        Скидка по картам
        '''
        values = struct.unpack('=BBdHH', tmcFile.read(14))

    def parseTKod21(self, tmcFile):
        '''
        Скидка по картам - оптовая
        '''
        values = struct.unpack('=BBdHHf', tmcFile.read(18))

    def parseTKod22(self, tmcFile):
        '''
        Скидка по картам - блочная
        '''
        values = struct.unpack('=BBdHHf', tmcFile.read(18))

    def parseTKod23(self, tmcFile):
        '''
        Скидка по картам - комплектная
        '''
        values = struct.unpack('=BBdHHfL', tmcFile.read(22))

    def parseTKod24(self, tmcFile):
        '''
        Скидка по картам - ассортиментная
        '''
        values = struct.unpack('=BBdHHfL', tmcFile.read(22))

    def parseTKod25(self, tmcFile):
        '''
        Скидка по картам - оптовая
        '''
        values = struct.unpack('=BBdHHL', tmcFile.read(18))

    def parseTKod30(self, tmcFile):
        '''
        Акция
        '''
        values = struct.unpack('=LLBdH', tmcFile.read(19))

    def parseTKod31(self, tmcFile):
        '''
        Акция оптовая
        '''
        values = struct.unpack('=LLBdHf', tmcFile.read(23))

    def parseTKod32(self, tmcFile):
        '''
        Акция блочная
        '''
        values = struct.unpack('=LLBdHf', tmcFile.read(23))

    def parseTKod33(self, tmcFile):
        '''
        Акция комплектная
        '''
        values = struct.unpack('=LLBdHfL', tmcFile.read(27))

    def parseTKod34(self, tmcFile):
        '''
        Акция ассортиментная
        '''
        values = struct.unpack('=LLBdHfL', tmcFile.read(27))

    def parseTKod35(self, tmcFile):
        '''
        Акция
        '''
        values = struct.unpack('=LLBdHf', tmcFile.read(23))

    def parseTKod50(self, tmcFile):
        '''
        Комплект
        '''
        values = struct.unpack('=LB', tmcFile.read(5))
        name = trimString(tmcFile.read(values[1]).decode('cp1251').encode('utf8'))

    def parseBarcodeBrcFormat(self, barcodeFile):
        '''
        Функция получения данных о штрихкоде
        Возвращает код товара, количество в штрихкоде и штрихкод
        '''
        values = struct.unpack('=LHB', barcodeFile.read(7))
        barcode = trimString(barcodeFile.read(values[2]).decode('cp1251').encode('utf8'))
        return str(values[0]), values[1], barcode.decode()

    def parseBarcodeBarFormat(self, barcodeFile):
        '''
        Функция получения данных о штрихкоде
        Возвращает код товара, количество в штрихкоде и штрихкод
        '''
        values = struct.unpack('=LHB', barcodeFile.read(7))
        barcode = trimString(barcodeFile.read(13).decode('cp1251').encode('utf8'))
        return str(values[0]), values[1], barcode.decode()

    def getTaxCode(self, percent):
        '''
        Поиск кода налога с процентом percent
        Если налог не найден, то создаёт объект налога и возвращает его код
        '''
        for tax in self.taxes:
            if abs(tax['lvat']['rate'] - percent) < 0.005:
                return tax['lvat']['code']
        code = len(self.taxes) + 1
        self.taxes.append(aif.LVat().addCmd(code=code, name='', rate=percent))
        return code

def general_method(workDirectory, data, node):
    log = logger.getUploadLogger("Node", node.id)
    result = False
    global errorList
    errorList = []
    try:
        result = Converter(workDirectory, data, node).run()
    except InserterError:
        raise
    except Exception as e:
        for errorLine in e.args:
            errorList.append(errorLine)
            log.error(errorLine)
    return result, errorList

def invalidObjectsCount():
    return errorObjectsCount
