# -*- coding: UTF-8 -*-
import itertools
import traceback
from itertools import groupby

from common import logger
from common import aif_objects as dicts
from upload.converter.smagPlugins.xml.artix_mapper.ukm_tmc_to_artix_tmc_mapper import UkmToArtixTmcMapper
from upload.converter.smagPlugins.xml.artix_mapper.enrichment_tmc_restrictions_mapper import EnrichmentTmcRestrictionsMapper

"""
Вспомогательный класс для получения команд на загрузку aif-объектов
"""


class ConverterUkm(object):

    def __init__(self, node_id, tnvedMapper={}, measureMapper={}, quantityOptionsMapper={}, tmcRestrictionsMapper={}):
        self.commands = {
            "updateItems": lambda xml_tag: self.tmc_handler(**xml_tag),
            "storePrices": lambda xml_tag: self.price_handler(**xml_tag),
            "storePricesAdd": lambda xml_tag: self.tprice_handler(**xml_tag),
            "taxes": lambda xml_tag: self.tax_handler(**xml_tag),
            "alcocode": lambda xml_tag: self.alcocode_handler(**xml_tag),
            "classif": lambda xml_tag: self.group_handler(**xml_tag),
            "kit": lambda xml_tag: self.kit_handler(**xml_tag),
            "storeCashiers": lambda xml_tag: self.cashiers_handler(**xml_tag),
            "picklists": lambda xml_tag: self.picklist_handler(**xml_tag)
        }
        self.aif_objects = {
            "updateItems": [dicts.Invent(), dicts.AlcoCode(), dicts.TmcAlcocode()],
            "storePrices": dicts.Price(),
            "storePricesAdd": dicts.AdditionalPrice(),
            "taxes": dicts.TaxGroup(),
            "classif": dicts.InventGroup(),
            "kit": [dicts.TmcKit(), dicts.TmcKitItem()],
            "storeCashiers": dicts.MCashUser(),
            "picklists": dicts.Picklist()
        }
        self.tnvedMapper = tnvedMapper
        self.measureMapper = measureMapper
        self.quantityOptionsMapper = quantityOptionsMapper
        self.log = logger.getUploadLogger("Node", node_id)
        # группы товаров имеют иерархическую структуру и надо настройки из tmcRestrictionsMapper распространить по иерархии групп,
        # потому что товарам назначается конечная группа, а настройки ограничений могут находиться в родительских группах
        try:
            EnrichmentTmcRestrictionsMapper(tmcRestrictionsMapper, node_id).enrich()
            UkmToArtixTmcMapper.set_mappers(self.tnvedMapper, self.measureMapper, self.quantityOptionsMapper,
                                            tmcRestrictionsMapper)
        except Exception as e:
            self.log.warn(str(e) + u' Не удалось обогатить группы товаров настройками')
            self.log.warn(traceback.format_exc())

    def get_command(self, line, data, kwargs=None):
        if kwargs is None:
            kwargs = {}
        kwargs['data'] = data
        return self.commands[line](kwargs)

    def get_clear_command(self, tag):
        aif_obj = self.aif_objects.get(tag)
        if aif_obj is not None:
            if type(aif_obj) == list:
                return [aif.clearCmd() for aif in list(aif_obj)]
            else:
                return [aif_obj.clearCmd()]
        return []

    def tmc_handler(self, data, tnveds):
        commands = []
        tmcs = UkmToArtixTmcMapper.map_tmc(data, tnveds)
        commands.extend([dicts.Invent().addCmd(**tmc) for tmc in tmcs])
        return commands

    def alcocode_handler(self, data):
        commands = []
        alcocodes_gen = UkmToArtixTmcMapper.map_alcocode(data)
        alcocodes, tmc_alcocodes = itertools.tee(alcocodes_gen)
        # на основе полученных алкокодов формируем команды на добавление в таблицы alcocode и tmcalcocode
        add_alcocodes = [dicts.AlcoCode().addCmd(code["alcocode"]) for code in alcocodes]  # type:list
        add_tmc_alcocodes = [dicts.TmcAlcocode().addCmd(**code) for code in tmc_alcocodes]  # type:list
        add_alcocodes.extend(add_tmc_alcocodes)
        commands.extend(add_alcocodes)
        return commands

    def group_handler(self, data):
        commands = []
        groups = UkmToArtixTmcMapper.map_groups(data)
        commands.extend([dicts.InventGroup().addCmd(**group) for group in groups])
        return commands

    def kit_handler(self, data):
        commands = []
        kits, kit_items = UkmToArtixTmcMapper.map_kit(data)
        # формируем команды на добавление в таблицы tmckit и tmckititem
        kit_add_cmds = [dicts.TmcKit().addCmd(**kit) for kit in kits]  # type: list
        kit_items_add_cmd = [dicts.TmcKitItem().addCmd(**kit_item) for kit_item in kit_items]  # type: list
        kit_add_cmds.extend(kit_items_add_cmd)
        commands.extend(kit_add_cmds)
        return commands

    def price_handler(self, data):
        commands = []
        prices = UkmToArtixTmcMapper.map_prices(data)
        commands.extend([dicts.Price().addCmd(**p) for p in prices])
        return commands

    def tprice_handler(self, data):
        commands = []
        tprices = UkmToArtixTmcMapper.map_tprices(data)
        commands.extend([dicts.AdditionalPrice().addCmd(**p) for p in tprices])
        return commands

    def tax_handler(self, data):
        commands = []
        taxes = UkmToArtixTmcMapper.map_tax(data)
        # получаем список вида taxgroupid -> [taxes]
        mapped_taxes = groupby(taxes, lambda x: x["code"])
        commands.extend([dicts.TaxGroup().addCmd(idtaxgroup=taxgroupid,
                                                 taxes=[{'name': tax["name"], 'rate': tax["rate"]} for tax in
                                                        taxes])
                         for taxgroupid, taxes in mapped_taxes])
        return commands

    def cashiers_handler(self, data):
        commands = []
        cashiers = UkmToArtixTmcMapper.map_cashier(data)
        commands.extend([dicts.MCashUser().addCmd(**user) for user in cashiers])
        return commands  # type: list

    def picklist_handler(self, data):
        commands = []
        picklists = UkmToArtixTmcMapper.map_picklist(data)
        commands.extend([dicts.Picklist().addCmd(**picklist) for picklist in picklists])
        # формируем команды на добавление в таблицу picklist
        return commands  # type: list
